/*
 * cellfs.h - CELLFS header file.
 *
 * Copyright (C) 2007-2012 Nippon Telegraph and Telephone Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * Written by Koji Sato.
 *
 * Maintained by Ryusuke Konishi <konishi.ryusuke@lab.ntt.co.jp> from 2008.
 */

#ifndef CELLFS_H
#define CELLFS_H

#if !defined(BINDGEN)
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/ioctl.h>
#endif /* BINDGEN */

#include <semaphore.h>
#include <linux/types.h>
#include <endian.h>
#include <byteswap.h>


/* FIX ME */
#ifndef __bitwise
typedef __u16	__le16;
typedef __u32	__le32;
typedef __u64	__le64;
typedef __u16	__be16;
typedef __u32	__be32;
typedef __u64	__be64;
#endif

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define le16_to_cpu(x)	((__u16)(x))
#define le32_to_cpu(x)	((__u32)(x))
#define le64_to_cpu(x)	((__u64)(x))
#define cpu_to_le16(x)	((__u16)(x))
#define cpu_to_le32(x)	((__u32)(x))
#define cpu_to_le64(x)	((__u64)(x))
#define be16_to_cpu(x)	bswap_16(x)
#define be32_to_cpu(x)	bswap_32(x)
#define be64_to_cpu(x)	bswap_64(x)
#define cpu_to_be16(x)	bswap_16(x)
#define cpu_to_be32(x)	bswap_32(x)
#define cpu_to_be64(x)	bswap_64(x)
#elif __BYTE_ORDER == __BIG_ENDIAN
#define le16_to_cpu(x)	bswap_16(x)
#define le32_to_cpu(x)	bswap_32(x)
#define le64_to_cpu(x)	bswap_64(x)
#define cpu_to_le16(x)	bswap_16(x)
#define cpu_to_le32(x)	bswap_32(x)
#define cpu_to_le64(x)	bswap_64(x)
#define be16_to_cpu(x)	((__u16)(x))
#define be32_to_cpu(x)	((__u32)(x))
#define be64_to_cpu(x)	((__u64)(x))
#define cpu_to_be16(x)	((__u16)(x))
#define cpu_to_be32(x)	((__u32)(x))
#define cpu_to_be64(x)	((__u64)(x))
#else
#error "unknown endian"
#endif	/* __BYTE_ORDER */

#ifndef BUG
#define BUG()	abort()
#endif

#include "cellfs_fs.h"

/* XXX: sector_t is not defined in user land */
typedef __u64 sector_t;	/* XXX: __u64 ?? */
typedef sector_t cellfs_blkoff_t;
typedef __u64 cellfs_cno_t;

#define CELLFS_FSTYPE	"cellfs"

#define CELLFS_CNO_MIN	((cellfs_cno_t)1)
#define CELLFS_CNO_MAX	(~(cellfs_cno_t)0)

#define CELLFS_SB_LABEL			0x0001
#define CELLFS_SB_UUID			0x0002
#define CELLFS_SB_FEATURES		0x0004
#define CELLFS_SB_COMMIT_INTERVAL	0x4000
#define CELLFS_SB_BLOCK_MAX		0x8000

/**
 * struct cellfs - cellfs object
 * @n_sb: superblock
 * @n_dev: device file
 * @n_ioc: ioctl file
 * @n_devfd: file descriptor of device file
 * @n_iocfd: file descriptor of ioctl file
 * @n_opts: options
 * @n_mincno: the minimum of valid checkpoint numbers
 * @n_sems: array of semaphores
 *     sems[0] protects garbage collection process
 */
struct cellfs {
	struct cellfs_super_block *n_sb;
	char *n_dev;
	/* char *n_mnt; */
	char *n_ioc;
	int n_devfd;
	int n_iocfd;
	int n_opts;
	cellfs_cno_t n_mincno;
	sem_t *n_sems[1];
};

#define CELLFS_OPEN_RAW		0x0001	/* Open RAW device */
#define CELLFS_OPEN_RDONLY	0x0002	/* Open CELLFS API in read only mode */
#define CELLFS_OPEN_WRONLY	0x0004	/* Open CELLFS API in write only mode */
#define CELLFS_OPEN_RDWR		0x0008	/* Open CELLFS API in read/write mode */
#define CELLFS_OPEN_GCLK		0x1000	/* Open GC lock primitive */

#define CELLFS_OPT_MMAP		0x01
#define CELLFS_OPT_SET_SUINFO	0x02


struct cellfs *cellfs_open(const char *, const char *, int);
void cellfs_close(struct cellfs *);

const char *cellfs_get_dev(const struct cellfs *);

void cellfs_opt_clear_mmap(struct cellfs *);
int cellfs_opt_set_mmap(struct cellfs *);
int cellfs_opt_test_mmap(struct cellfs *);

void cellfs_opt_clear_set_suinfo(struct cellfs *);
int cellfs_opt_set_set_suinfo(struct cellfs *);
int cellfs_opt_test_set_suinfo(struct cellfs *);

cellfs_cno_t cellfs_get_oldest_cno(struct cellfs *);

struct cellfs_super_block *cellfs_get_sb(struct cellfs *);


#define CELLFS_LOCK_FNS(name, index)					\
static inline int cellfs_lock_##name(struct cellfs *cellfs)		\
{									\
	return sem_wait(cellfs->n_sems[index]);				\
}									\
static inline int cellfs_trylock_##name(struct cellfs *cellfs)		\
{									\
	return sem_trywait(cellfs->n_sems[index]);			\
}									\
static inline int cellfs_unlock_##name(struct cellfs *cellfs)		\
{									\
	return sem_post(cellfs->n_sems[index]);				\
}

CELLFS_LOCK_FNS(cleaner, 0)

/**
 * struct cellfs_psegment - partial segment iterator
 * @p_segnum: segment number
 * @p_blocknr: block number of partial segment
 * @p_segblocknr: block number of segment
 * @p_nblocks: number of blocks in segment
 * @p_maxblocks: maximum number of blocks in segment
 * @p_blksize: block size
 * @p_seed: CRC seed
 */
struct cellfs_psegment {
	struct cellfs_segment_summary *p_segsum;
	sector_t p_blocknr;

	sector_t p_segblocknr;
	size_t p_nblocks;
	size_t p_maxblocks;
	size_t p_blksize;
	__u32 p_seed;
};

/**
 * struct cellfs_file - file iterator
 * @f_finfo: file information
 * @f_blocknr: block number
 * @f_offset: byte offset from the beginning of segment
 * @f_index: index
 * @f_psegment: partial segment
 */
struct cellfs_file {
	struct cellfs_finfo *f_finfo;
	sector_t f_blocknr;

	unsigned long f_offset;
	unsigned int f_index;
	const struct cellfs_psegment *f_psegment;
};

/**
 * struct cellfs_block - block iterator
 * @b_binfo: block information
 * @b_blocknr: block number
 * @b_offset: byte offset from the beginning of segment
 * @b_index: index
 * @b_dsize: size of data block information
 * @b_nsize: size of node block information
 * @b_file: file
 */
struct cellfs_block {
	void *b_binfo;
	sector_t b_blocknr;

	unsigned long b_offset;
	unsigned int b_index;
	size_t b_dsize;
	size_t b_nsize;
	const struct cellfs_file *b_file;
};

/* virtual block number and block offset */
#define CELLFS_BINFO_DATA_SIZE		(sizeof(__le64) + sizeof(__le64))
/* virtual block number */
#define CELLFS_BINFO_NODE_SIZE		sizeof(__le64)
/* block offset */
#define CELLFS_BINFO_DAT_DATA_SIZE	sizeof(__le64)
/* block offset and level */
#define CELLFS_BINFO_DAT_NODE_SIZE	(sizeof(__le64) + sizeof(__le64))


/* partial segment iterator */
void cellfs_psegment_init(struct cellfs_psegment *, __u64,
			 void *, size_t, const struct cellfs *);
int cellfs_psegment_is_end(const struct cellfs_psegment *);
void cellfs_psegment_next(struct cellfs_psegment *);

#define cellfs_psegment_for_each(pseg, segnum, seg, nblocks, cellfs)	\
	for (cellfs_psegment_init(pseg, segnum, seg, nblocks, cellfs);	\
	     !cellfs_psegment_is_end(pseg);				\
	     cellfs_psegment_next(pseg))

/* file iterator */
void cellfs_file_init(struct cellfs_file *, const struct cellfs_psegment *);
int cellfs_file_is_end(const struct cellfs_file *);
void cellfs_file_next(struct cellfs_file *);

static inline int cellfs_file_is_super(const struct cellfs_file *file)
{
	__u64 ino;

	ino = le64_to_cpu(file->f_finfo->fi_ino);
	return ino == CELLFS_DAT_INO;
}

#define cellfs_file_for_each(file, pseg)		\
	for (cellfs_file_init(file, pseg);	\
	     !cellfs_file_is_end(file);		\
	     cellfs_file_next(file))

/* block iterator */
void cellfs_block_init(struct cellfs_block *, const struct cellfs_file *);
int cellfs_block_is_end(const struct cellfs_block *);
void cellfs_block_next(struct cellfs_block *);

static inline int cellfs_block_is_data(const struct cellfs_block *blk)
{
	return blk->b_index < le32_to_cpu(blk->b_file->f_finfo->fi_ndatablk);
}

static inline int cellfs_block_is_node(const struct cellfs_block *blk)
{
	return blk->b_index >= le32_to_cpu(blk->b_file->f_finfo->fi_ndatablk);
}

#define cellfs_block_for_each(blk, file)		\
	for (cellfs_block_init(blk, file);	\
	     !cellfs_block_is_end(blk);		\
	     cellfs_block_next(blk))

#define CELLFS_SB_BLOCK_SIZE_SHIFT	10

struct cellfs_super_block *cellfs_sb_read(int devfd);
int cellfs_sb_write(int devfd, struct cellfs_super_block *sbp, int mask);

int cellfs_read_sb(struct cellfs *);

ssize_t cellfs_get_segment(struct cellfs *, unsigned long, void **);
int cellfs_put_segment(struct cellfs *, void *);
size_t cellfs_get_block_size(struct cellfs *);
__u64 cellfs_get_segment_seqnum(const struct cellfs *, void *, __u64);
__u64 cellfs_get_reserved_segments(const struct cellfs *cellfs);

int cellfs_change_cpmode(struct cellfs *, cellfs_cno_t, int);
ssize_t cellfs_get_cpinfo(struct cellfs *, cellfs_cno_t, int,
			 struct cellfs_cpinfo *, size_t);
int cellfs_delete_checkpoint(struct cellfs *, cellfs_cno_t);
int cellfs_get_cpstat(const struct cellfs *, struct cellfs_cpstat *);
ssize_t cellfs_get_suinfo(const struct cellfs *, __u64, struct cellfs_suinfo *,
			 size_t);
int cellfs_set_suinfo(const struct cellfs *, struct cellfs_suinfo_update *,
		     size_t);
int cellfs_get_sustat(const struct cellfs *, struct cellfs_sustat *);
ssize_t cellfs_get_vinfo(const struct cellfs *, struct cellfs_vinfo *, size_t);
ssize_t cellfs_get_bdescs(const struct cellfs *, struct cellfs_bdesc *, size_t);
int cellfs_clean_segments(struct cellfs *, struct cellfs_vdesc *, size_t,
			 struct cellfs_period *, size_t, __u64 *, size_t,
			 struct cellfs_bdesc *, size_t, __u64 *, size_t);
int cellfs_sync(const struct cellfs *, cellfs_cno_t *);
int cellfs_resize(struct cellfs *cellfs, off_t size);
int cellfs_set_alloc_range(struct cellfs *cellfs, off_t start, off_t end);

static inline __u64 cellfs_get_nsegments(const struct cellfs *cellfs)
{
	return le64_to_cpu(cellfs->n_sb->s_nsegments);
}

static inline __u32 cellfs_get_blocks_per_segment(const struct cellfs *cellfs)
{
	return le32_to_cpu(cellfs->n_sb->s_blocks_per_segment);
}

#endif	/* CELLFS_H */
