/*
 * cellfs_cleaner.h - CELLFS cleaner controller routines
 *
 * Licensed under LGPLv2: the complete text of the GNU Lesser General
 * Public License can be found in COPYING file of the cellfs-utils
 * package.
 *
 * Copyright (C) 2007-2012 Nippon Telegraph and Telephone Corporation.
 */

#ifndef CELLFS_CLEANER_H
#define CELLFS_CLEANER_H

#ifndef BINDGEN
#include <sys/types.h>
#endif
#include <stdint.h>

#include "cellfs.h"

struct cellfs_cleaner;

#define CELLFS_CLEANER_OPEN_GCPID	(1 << 0)
#define CELLFS_CLEANER_OPEN_QUEUE	(1 << 1)

struct cellfs_cleaner *cellfs_cleaner_launch(const char *device,
					   const char *mntdir,
					   unsigned long protperiod);
struct cellfs_cleaner *cellfs_cleaner_open(const char *device,
					 const char *mntdir, int oflag);

int cellfs_cleaner_ping(struct cellfs_cleaner *cleaner);

pid_t cellfs_cleaner_pid(const struct cellfs_cleaner *cleaner);
const char *cellfs_cleaner_device(const struct cellfs_cleaner *cleaner);

void cellfs_cleaner_close(struct cellfs_cleaner *cleaner);

/* cleaner command arguments */
struct cellfs_cleaner_args {
	uint16_t valid;
	uint16_t npasses; /* number of passes */
	uint16_t usage_rate_threshold;
	uint16_t nsegments_per_clean;
	uint8_t pad;
	uint8_t min_reclaimable_blocks_unit;
	uint16_t cleaning_interval;
	uint32_t cleaning_interval_nsec;
	uint64_t protection_period; /* protection period in seconds */
	uint64_t start_segnum;	/* start segment number */
	uint64_t nsegs;		/* number of segments */
	uint32_t runtime; /* runtime in seconds */
	uint32_t min_reclaimable_blocks;
};

enum cellfs_cleaner_args_unit {
	CELLFS_CLEANER_ARG_UNIT_NONE = 0,
	CELLFS_CLEANER_ARG_UNIT_PERCENT,
	CELLFS_CLEANER_ARG_UNIT_KB,     /* kilo-byte (kB) */
	CELLFS_CLEANER_ARG_UNIT_KIB,    /* kibi-byte (KiB) */
	CELLFS_CLEANER_ARG_UNIT_MB,     /* mega-byte (MB) */
	CELLFS_CLEANER_ARG_UNIT_MIB,    /* mebi-byte (MiB) */
	CELLFS_CLEANER_ARG_UNIT_GB,     /* giga-byte (GB) */
	CELLFS_CLEANER_ARG_UNIT_GIB,    /* gibi-byte (GiB) */
	CELLFS_CLEANER_ARG_UNIT_TB,     /* tera-byte (TB) */
	CELLFS_CLEANER_ARG_UNIT_TIB,    /* tebi-byte (TiB) */
	CELLFS_CLEANER_ARG_UNIT_PB,     /* peta-byte (PB) */
	CELLFS_CLEANER_ARG_UNIT_PIB,    /* pebi-byte (PiB) */
	CELLFS_CLEANER_ARG_UNIT_EB,     /* exa-byte (EB) */
	CELLFS_CLEANER_ARG_UNIT_EIB,    /* exbi-byte (EiB) */

	CELLFS_CLEANER_ARG_MIN_BINARY_SUFFIX = CELLFS_CLEANER_ARG_UNIT_KB,
	CELLFS_CLEANER_ARG_MAX_BINARY_SUFFIX = CELLFS_CLEANER_ARG_UNIT_EIB,
};

/* valid flags */
#define CELLFS_CLEANER_ARG_PROTECTION_PERIOD		(1 << 0)
#define CELLFS_CLEANER_ARG_NSEGMENTS_PER_CLEAN		(1 << 1)
#define CELLFS_CLEANER_ARG_CLEANING_INTERVAL		(1 << 2)
#define CELLFS_CLEANER_ARG_USAGE_RATE_THRESHOLD		(1 << 3) /* reserved */
#define CELLFS_CLEANER_ARG_START_SEGNUM			(1 << 4) /* reserved */
#define CELLFS_CLEANER_ARG_NSEGS				(1 << 5) /* reserved */
#define CELLFS_CLEANER_ARG_NPASSES			(1 << 6) /* reserved */
#define CELLFS_CLEANER_ARG_RUNTIME			(1 << 7) /* reserved */
#define CELLFS_CLEANER_ARG_MIN_RECLAIMABLE_BLOCKS	(1 << 8)

enum {
	CELLFS_CLEANER_STATUS_IDLE,
	CELLFS_CLEANER_STATUS_RUNNING,
	CELLFS_CLEANER_STATUS_SUSPENDED,
};

int cellfs_cleaner_get_status(struct cellfs_cleaner *cleaner, int *status);
int cellfs_cleaner_run(struct cellfs_cleaner *cleaner,
		      const struct cellfs_cleaner_args *args, uint32_t *jobid);
int cellfs_cleaner_suspend(struct cellfs_cleaner *cleaner);
int cellfs_cleaner_resume(struct cellfs_cleaner *cleaner);
int cellfs_cleaner_tune(struct cellfs_cleaner *cleaner,
		       const struct cellfs_cleaner_args *args);
int cellfs_cleaner_reload(struct cellfs_cleaner *cleaner, const char *conffile);
int cellfs_cleaner_wait(struct cellfs_cleaner *cleaner, uint32_t jobid,
		       const struct timespec *abs_timeout);
int cellfs_cleaner_wait_r(struct cellfs_cleaner *cleaner, uint32_t jobid,
			 const struct timespec *timeout);
int cellfs_cleaner_stop(struct cellfs_cleaner *cleaner);
int cellfs_cleaner_shutdown(struct cellfs_cleaner *cleaner);

extern void (*cellfs_cleaner_logger)(int priority, const char *fmt, ...);
extern void (*cellfs_cleaner_printf)(const char *fmt, ...);
extern void (*cellfs_cleaner_flush)(void);

#endif /* CELLFS_CLEANER_H */
